// utility verbs for producing rasters (from vector field and density plots)
// and rendering them to PBM formats.

#include <iostream>
#include <sstream>
#include <iomanip>
#include <cstdlib>
#include "jarray.h"

typedef double Real;
typedef double Int;

extern bool PROGRESS;

using namespace std;

// returns interpolated over X and Y components value from the grid, taken
// at a specific value of z index (0<=iz<Nz). Coordinates x and y should assume
// values in ranges 0<x<Lx, 0<y<Ly or they will be corrected.
Real interpolate(Int Lx, Int Ly, Real x, Real y, const jarray_of_type<Real> &V) {
  assert (V.rank()==2);
  const int Nx=V.extent(0);
  const int Ny=V.extent(1);
  
  // check and forgivingly pull back in range
  if (x<0) x=0; if (x>Lx) x=Lx;
  if (y<0) y=0; if (y>Ly) y=Ly;

  Real rx=x/Lx*(Nx-1), ry=y/Ly*(Ny-1);   // put on grid
  // a simple bilinear interpolation
  // In simulation we assume that ix, iy and iz specify the center
  // of the element. In visualization we assume the grid is shifted by 
  // half-step in each direction. It is simpler this way.
  // The conditionals below guard against the case of the coordinate
  // landing exactly on the grid node. However it seems improbable, this
  // does happen.
  Real f1= (ceil(rx)==floor(rx))? V(floor(rx),floor(ry)) :
    ( ceil(rx)-rx)*V(floor(rx),floor(ry))+
    (rx-floor(rx))*V( ceil(rx),floor(ry));
  Real f2= (ceil(rx)==floor(rx))? V(floor(rx),ceil(ry)) :
    ( ceil(rx)-rx)*V(floor(rx), ceil(ry))+
    (rx-floor(rx))*V( ceil(rx), ceil(ry));
  return (ceil(ry)==floor(ry))? f1 : ( ceil(ry)-ry)*f1 + (ry-floor(ry))*f2;
};

// Interpolates scalar and plots vector fields.
// The right argument can be either 2d array (shape N x M) or 2d array of 2-d 
// vectors (shape  2 x N x M). The left argument, if unboxed, is a length-two 
// integer list specifying dimension of the resulting
// raster (which can be larger or smaller than the data dimension). Boxed left arguments 
// (which will allow to specify additional rendering options) are not yet 
// implemented.
// Plotting vectors, results in grayscale rasters (represented by double arrays of
// shape N M, where N M are given as left argument to rasterplot; their elements take 
// values in range 0.0-1.0).
jarray rasterplot(jarray x, jarray y) {
  assert(x.getEngine()==y.getEngine());
  jengine *je=y.getEngine();
  jarray_of_type<Int> xt(x);
  jarray_of_type<Real> yt(y);
  assert((x.rank()==1) && (x.size()==2));
  assert((y.rank()==2)||((y.rank()==3)&&(y.extent(2)==2)));
  Int Lx=xt[0],Ly=xt[1];
  if (y.rank()==2) { // interpolate
    jarray_of_type<Real> r(je,Lx,Ly);
    for (int i=0;i<Lx;i++)
      for (int j=0;j<Ly;j++) {
	r(i,j)=interpolate(Lx,Ly,i,j,yt);
      };
    return r;
  } else { // vector field plot
    // we use Line Integral Convolution (LIC) method
    const int Nx=y.extent(0);
    const int Ny=y.extent(1);

    // TODO: this copy should be avoided
    jarray_of_type<Real> vx(je, Nx,Ny), vy(je, Nx,Ny);
    for (int i=0;i<Nx;i++)
      for (int j=0;j<Ny;j++) {
	vx(i,j)=yt(i,j,0);
	vy(i,j)=yt(i,j,1);
      };
    
    const int l=10; // vector length in pixels
    srand (1); // rewind RNG

    // create random permutation of image pixels
    int order[(int)(Lx*Ly)];
    for (int i=0; i<Lx*Ly; i++) order[i]=i;
    for (int i=0; i<Lx*Ly; i++) {
      int k=rand() % (int)(Lx*Ly);
      int t = order[i];
      order[i] = order[k];
      order[k]=t;
    };

    // create initial image, filled with white noise
    jarray_of_type<Real> image(je, Lx, Ly);
    for (int i=0; i<Lx; i++)
      for (int j=0; j<Ly; j++)
	image(i,j)=((double)rand()/(double)RAND_MAX);
    
    // traverse every pixel
    int pct=0;
    if (PROGRESS) cerr << "Drawing vector field:     ";
    for (int i=0; i<Lx*Ly; i++) {
      // coordinates of the pixel
      int ix=order[i]%(int)Lx;
      int iy=order[i]/(int)Lx;
      Real avg=image(ix,iy);         // color of the pixel (seeds an average)
      int navg=1;                      // number of averaged points
    
      Real ss[]={0.3*Nx/Lx,-0.3*Ny/Ly}; // step sizes for two directions to traverse
      for (int m=0;m<2;m++) {
	int ox,cx, oy, cy;
	ox=cx=ix; oy=cy=iy;              // initial point in the image
	Real cR[2]={cx,cy};  // initial point in the field
	for (int k=0;k<l;k++) { // travel l pixels
	  int nsteps=0;
	  while ((cx==ox)&&(cy==oy)&&(nsteps<1000)) { // integrate until we move
	    // perform Runge-Kutta step
	    Real tmpR[2];
	    Real A[2];
	    A[0]=interpolate(Lx,Ly,cR[0],cR[1],vx);
	    A[1]=interpolate(Lx,Ly,cR[0],cR[1],vy);
	    for(int j=0;j<2;j++) tmpR[j]=cR[j]+A[j]*(ss[m]/2.0);
	    Real B[2];
	    B[0]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vx);
	    B[1]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vy);
	    for(int j=0;j<2;j++) tmpR[j]=cR[j]+B[j]*(ss[m]/2.0);
	    Real C[2];
	    C[0]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vx);
	    C[1]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vy);
	    for(int j=0;j<2;j++) tmpR[j]=cR[j]+C[j]*(ss[m]);
	    Real D[2];
	    D[0]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vx);
	    D[1]=interpolate(Lx,Ly,tmpR[0],tmpR[1],vy);
	    for(int j=0;j<2;j++) cR[j]+=(A[j]+2*B[j]+2*C[j]+D[j])/6.0*ss[m];
	    // check that we're not stuck
	    const Real eps=0.00001;
	    if ((abs(A[0])<eps)&&(abs(A[1])<eps)) break;
	    // next pixel
	    cx=round(cR[0]);
	    cy=round(cR[1]);
	    nsteps++;
	  };
	  // stop if we moved out of image
	  if ((cx<0)||(cx>(Lx-1))||(cy<0)||(cy>(Ly-1))) break;
	  avg+=image(cx,cy);
	  navg++;
	  ox=cx; oy=cy;
	};
      };
      // average the color
      avg/=navg;
      assert((ix>=0)&&(ix<Lx));
      assert((iy>=0)&&(iy<Ly));      
      image(ix,iy)=avg;
      
      if (PROGRESS) {
	int npct=(int) (i*100.0/(Lx*Ly));
	if (npct!=pct) {
	  cerr << "\b\b\b\b" << setw(3) << right << npct << "%";
	  pct=npct;
	};
      };
    };
    
    if (PROGRESS)
      cerr << "\b\b\b\b" << "100%" << endl;
    
    return image;
  };
};

/*
  rainbow =:_100 0 0 1 , _1 0 0 1 , _0.5 0 1 1 , 0 0 1 0 , 0.5 1 1 0, 1.0 1 0 0,: 100 1 0 0
  range =: [ {~ [: (, <:) {.@|:@[ I. ]
  interpolate =: [: +/ [: _3&{."1 ] * 1 - [: (% +/) [ |@- {."1@:]
  cmap =: ] interpolate range
  
 */

// renders y in PNM, returned in a literal array
// the left argument specifies the number of colors (or grayscales)
// the shape of right argument must be N x M (for grayscale images) 
// or N x M x 3 for RGB color images. All color values in the supplied
// rasters must be in range 0.0 - 1.0 .
jarray pnmplot(jarray x, jarray y) {
  assert(x.getEngine()==y.getEngine());
  jengine *je=y.getEngine();
  jarray_of_type<Int> xt(x);
  assert(x.size()==1);
  int colors=xt[0];
  jarray_of_type<Real> yt(y);
  assert((y.rank()==2)||((y.rank()==3)&&(y.extent(2)==3)));
  int Nx=y.extent(0);
  int Ny=y.extent(1);
  bool gray=(y.rank()==2);
  stringstream out;
  if (gray) out << "P2"; else out << "P3";
  out << endl << Nx << ' ' << Ny << endl;
  out << colors << endl;
  for (int j=0; j<Ny; j++) 
    for (int i=0; i<Nx; i++) {
      if (gray)
	out << (int)round(colors*yt(i,j)) << ' ';
      else
	out << (int)round(colors*yt(i,j,0)) << ' '
	    << (int)round(colors*yt(i,j,1)) << ' '
	    << (int)round(colors*yt(i,j,2)) << ' ';
    };
  out << endl;
  jarray res(je,out.str());
  return res;
};

jarray pnmplot(jarray y) {
  jengine *je=y.getEngine();
  jarray_of_type<Real> xt(je);
  xt[0]=255;
  return pnmplot(xt,y);
};
