#include <cassert>
#include "trjfile.h"
#include "sys/types.h"

using namespace std;

trjfile::trjfile():neq(0) {};

void trjfile::setHeader(int pid, int pneq) {
  assert(neq==0); // only on fresh file
  int32_t id=pid;  // reserved
  int32_t tneq=neq=pneq;
  trj.seekg(0, std::ios::end);
  streampos file_size = trj.tellg();
  // we can modify the header only before the first frame is written
  assert(file_size<=header_size()); 
  trj.seekg(0);
  // TODO: convert to little-endian on big-endian platforms
  trj.write((char*)&id,sizeof(id));
  trj.write((char*)&tneq,sizeof(tneq));
  trj.flush();
};

bool trjfile::open(std::string trjname, bool create) {
  if (trj.is_open()) {
    trj.close();
  };
  if (create) { // create trajectory file if it wasn't there
    trj.open(trjname.c_str(), ios::out | ios::binary | ios::app);
    trj.flush();
    trj.close();
  };
  trj.open(trjname.c_str(), ios::in | ios::out | ios::binary);
  if (trj.is_open()) {
    // read header
    int32_t id=0;  // reserved
    int32_t tneq=0;
    // TODO: convert from little-endian on big-endian platforms
    trj.read((char*)&id,sizeof(id));
    trj.read((char*)&tneq,sizeof(tneq)); neq=tneq;
    if (!trj.good()) {
      if (!create) {
	trj.close(); 
	return false;
      };
      trj.clear();
      setHeader(0,0); // reserve space for the header
    };
    return true;
  } else return false;
};

bool trjfile::close() {
  if (!trj.is_open()) return false;
  trj.close();
  neq=0;
  return true;
};

std::streampos trjfile::header_size() {
  return 2*sizeof(int32_t);
};
  
/** frame size on disk */
std::streampos trjfile::frame_size() {
  assert(neq>0);
  return sizeof(double)+neq*sizeof(double);
};

bool trjfile::good() {
  bool file_ok = trj.is_open() && trj.good();
  if (!file_ok) return false;
  // check that file is properly positioned
  streampos curr=trj.tellg();
  if (curr<header_size()) return false;
  curr-=header_size();
  if (curr==0) return true;
  return (curr % frame_size())==0;
};

int trjfile::size() {
  streampos curr=trj.tellg();
  trj.seekg(0, std::ios::end);
  streampos file_size = trj.tellg();
  trj.seekg(curr, ios::beg);
  assert(file_size>=header_size());
  if (file_size==header_size()) return 0;
  return (file_size-header_size())/frame_size();
};

bool trjfile::toFrame(int iframe) {
  if (iframe>=0) {
    streampos curr=header_size()+frame_size()*iframe;
    trj.seekg(curr, ios::beg);
    return trj.good();
  } else {
    int nframes=size();
    if (nframes+iframe>0) { // load the -iframe'th from the end
      toFrame(nframes+iframe);
      return trj.good();
    } else return false;
  };
};

bool trjfile::hasNextFrame() {
  streampos curr=trj.tellg();
  trj.seekg(0, std::ios::end);
  streampos file_size = trj.tellg();
  trj.seekg(curr, ios::beg);
  if (file_size==curr) return false;
  return (file_size-curr)>=frame_size();
};

int trjfile::getNEQ() {
  return neq;
};

/** saves frame to the file */
bool trjfile::saveFrame(const int dneq, const double &T, const double* data) {
  if (!good()) return false;
  if (neq==0) // writing the first frame defines neq
    setHeader(0,dneq);
  assert(neq == dneq);
  // TODO: convert to little-endian on big-endian platforms
  trj.write((char *)&T,sizeof(double));
  trj.write((char *)data,neq*sizeof(double));
  trj.flush(); // sync the changes
  return true;
};

bool trjfile::loadFrame(double &T, double* data) {
  if (!good()) return false;
  assert(neq>0);
  // TODO: convert from little-endian on big-endian platforms
  trj.read((char*)&T,sizeof(double));
  trj.read((char*)data,neq*sizeof(double));
  return trj.good();
};
