#ifndef YACTS_H
#define YACTS_H

#include <istream>
#include <fstream>
#include <string>
#include "jplus.h"
#include "trjfile.h"

/**
 * YACTS -- yet another continuous time simulator. It uses J+ to
 * specify a system of ODEs and Sundials library to solve them.
 */
class yacts : public jplus, protected trjfile {
 public:
  /** Initializes YACTS */
  yacts();
  
  /** Load YACTS script. 
      The script must define the following variables:
      T, double scalar -- initial(first assignment), current and next frame time;
      S, double array  -- initial(first assignment) and current state;
      dSdT, double array -- time derivative of state;
      OUT, character vector -- textual rendering of the current state.
      @param script stream to load YACTS script from.
   */
  bool init(std::istream &script);
  
  /** Frees YACTS memory. */
  ~yacts();
  
  /** Initializes YACTS library functions and makes them avaliable to J. */
  virtual void libInit();

  /** Sets the name of output variable.
   * @param newOut new output variable name.
   * @return "true" if success (the name was OK) */
  bool setOut(std::string newOut);

  /** Enters read(stdin)-eval-print(stdout) loop until the end of stdin.
   * @return 0.*/
  int REPL();

  /** Computes trajectory filename base, using hash of "important" parts of
   *    yacts script.
   * @param prefix trajectory filename prefix.
   * @return trajectory filename base (without extension).
   */
  std::string getTrajectoryFilenameBase(std::string prefix);

  /** Initializes ODE solver and the state either from the start of the 
   * problem or from the last saved trajectory frame.
   * @param prefix trajectory filename prefix.
   * @return true upon success. */
  bool initTrajectory(std::string prefix);

  /** Returns the number of computed frames in the current trajectory file.
   * @return number of frames in the trajectory */
  int size();
  
  /** Sets specified frame as "current", loads it.
   * @param iframe index of the frame to become current. 
   * @return "true" upon success. */
  bool setFrame(int iframe);

  /** Returns true if the next frame is already in the trajectory file and does not
   *  have to be (and will not be) calculated. 
   * @return "true" if the next frame is in trajectory */
  bool hasNextFrameStored();

  /** Advances to the next trajectory frame by computing it (if necessary) and 
   *   saving (if computed).
   * @return "true" if all is ok. */
  bool nextFrame();
  
  /** Computes and returns the "OUT" variable, corresponding to the current frame. 
   * @return the computed output.*/
  std::string process();

  friend int runTests(); // testsuite is also our friend :-)

 private:
  void *cvode_mem;   // memory, allocated by sundials CVODE solver
  void *nvs;         // current state 
  std::string outName; // out variable name

  /** load frame into J+ environment from properly positioned trajectory file */
  bool loadFrame();
  /** save frame from J+ environment to properly positioned trajectory file */
  bool saveFrame();

};

#endif
